*                               S L A L O M
*
*    Scalable Language-independent Ames Laboratory One-minute Measurement
*
*     The following program is the first benchmark based on fixed time rather
*  than fixed problem comparison.  Not only is fixed time more representative
*  of the way people use computers, it also greatly increases the scope and
*  longevity of the benchmark.  SLALOM is very scalable, and can be used to
*  compare computers as slow as 152 floating-point operations per second
*  (FLOPS) to computers running a trillion times faster.  The scalability can
*  be used to compare single processors to massively parallel collections
*  of processors, and to study the space of problem size vs. ensemble size
*  in fine detail.  It resembles the LINPACK benchmark since it involves
*  factoring and backsolving a (nearly) dense 64-bit matrix, but incorporates
*  a number of improvements to that benchmark that we hope will make SLALOM
*  a better reflection of general system performance.
*
*     The SLALOM benchmark solves a complete, real problem (optical radiosity
*  on the interior of a box), not a contrived kernel or a synthetic mixture of
*  sample operations.  SLALOM is unusual since it times input, problem setup,
*  solution, and output, not just the solution.  For slower computers, the
*  problem setup will take the majority of the time; it grows as the square of
*  the problem size.  The solver grows as the cube of the problem size, and
*  dominates the time for large values of n.
*
*     While the following is Fortran 77, you are free to translate it into any
*  language you like, including assembly language specific to one computer.
*  You may use compiler directives, hand-tuned library calls, loop unrolling,
*  and even change the algorithm, if you can provide a convincing argument
*  that the program still works for the full range of possible inputs.  For
*  example, if you replace the direct solver with an iterative one, you must
*  make sure your method is correct even when the geometry is quite eccentric
*  and the box faces are highly reflective (rho = 0.999).
*
*     The "FixedT" driver should be used with the value of 60 seconds for the
*  SLALOM benchmark.  The work done for a particular problem size is figured
*  after timing has ceased, so there is no overhead for work assessment.  The
*  residual check ||Ax - b|| is also done after timing has ceased.  Two
*  computers are to be compared by their problem size n, not by MFLOPS or
*  execution times.  Times will always be near one minute in SLALOM.  For
*  continuity with other benchmarks, we estimate MFLOPS.  The MFLOPS counts
*  use the following weights for floating-point operations, based on
*  weights used by Lawrence Livermore National Laboratory:
*
*                        OPERATION                       WEIGHT
*                    a=b, a=(constant)                      0
*       a.LT.0, a.LE.0, a.EQ.0, a.NE.0, a.GT.0, a.GE.0      0
*                 -a, ABS(a), SIGN(a, b)                    0
*                   a+b, a-b, a*b, a**2                     1
*       a.LT.b, a.LE.b, a.EQ.b, a.NE.b, a.GT.b, a.GE.b      1
*           INT(a), FLOAT(i) (even when implicit)           1
*                         NINT(a)                           2
*                        1/a, -1/a                          3
*                           a/b                             4
*                          SQRT(a)                          4
*               Format to or from ASCII string              6
*       SIN(a), COS(a), TAN(a), LOG(a), ATN(a), EXP(a)      8
*
*     We invite you to share with us the results of any measurements that you
*  make with SLALOM.  We do NOT accept anonymous data; machine timings will be
*  referenced and dated.
*
*     The least you need to do to adapt SLALOM to your computer is:
*
*        1.  In the "Measure" routine, set NMAX to a value large enough to keep
*            the computer working for a minute.  Vary it slightly if it helps
*            (for reasons of cache size, interleaving, etc.)
*
*        2.  Replace the timer call in "When" with the most accurate wall-clock
*            timer at your disposal.  If only CPU time is available, try to run
*            the job standalone or at high priority, since we are ultimately
*            interested in the top of the statistical range of performance.
*
*        3.  Edit in the information specific to your test in the "What"
*            routine, so that final output will be automatically annotated.
*
*        4.  Compile, link, and run the program, interacting to select values
*            of n that bracket a time of one minute.  Once everything is
*            running, run it as a batch job so as to record the session.
*
*     Examples of ways you may optimize performance:
*
*        1.  Unroll the loops in SetUp1 and SetUp2; it is possible to
*            vectorize both SetUp1 and SetUp2 at the cost of some extra
*            operations, program complexity, and storage.
*
*        2.  Replace the innermost loops of Solver with calls to well-tuned
*            libraries of linear algebra routines, such as DDOT from the
*            Basic Linear Algebra Subroutines (level 1 BLAS).  Better still,
*            use a tuned library routine for all of Solver; the sparsity
*            exploited in Solver is only a few percent, so you will usually
*            gain more than you lose by applying a dense symmetric solver.
*
*        3.  Parallelize the SetUp and Solver routines; all are highly
*            parallel.  Each element of the matrix can be constructed
*            independently, once each processor knows the geometry and part of
*            the partitioning into regions.  A substantial body of literature
*            now exists for performing the types of operations in Solver in
*            parallel.
*
*        4.  Overlap computation with output.  Once the Region routine is done,
*            the first part of the output file (patch geometry) can be written
*            while the radiosities are being calculated.
*
*     Examples of what you may NOT do:
*
*        1.  The tuning must not be made specific to the particular input
*            provided.  For example, you may not eliminate IF tests simply
*            because they always come out the same way for this input; you
*            may not use precomputed answers or table look-up unless those
*            answers and tables cover the full range of possible inputs; and
*            you may not exploit symmetry for even values of the problem size.
*
*        2.  You may not disable the self-consistency tests in SetUp3 and
*            Verify, nor alter their tolerance constants.
*
*        3.  You may not change the input or output files to unformatted
*            binary or other format that would render them difficult to create
*            or read for humans.
*
*        4.  You may not eliminate the reading of the "geom" file by putting
*            its data directly into the compiled program.
*
*        5.  You may not change any of the work assessments in Meter.  If you
*            use more floating-point operations than indicated, you must still
*            use the assessments provided.  If you find a way to use fewer
*            operations and still get the job done for arbitrary input
*            parameters, please tell us!
*
*                             -John Gustafson, Diane Rover,
*                              Stephen Elbert, and Michael Carter
*                              Ames Laboratory, Ames, Iowa 50011
*
*******************************************************************************
*  The following program finds a value n such that a problem of size n        *
*  takes just under "goal" seconds to execute.                                *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, Ames, Iowa  10/17/90.                                     *
*                                                                             *
*  Calls:  Meter   Measures execution time for some application.              *
*          What    Prints work-timing statistics and system information.      *
*******************************************************************************
      PROGRAM FixedT
*
*  Variables used:
*    goal    User input, fixed-time benchmark goal, in seconds.
*    timing  Elapsed time returned by Meter routine, in seconds.
*    work    In this case, number of floating-point operations performed.
*    mean    Average between upper and lower bounds for bisection method.
*    n       The problem size.  Time and work should increase with n.
*    nupper  Upper bound on problem size, used in iterating toward goal.
*    ok      Flag indicating computational success of Meter.
*
      REAL*8 goal, timing, work
      INTEGER mean, n, nupper
      LOGICAL ok
*
*  Get desired number of seconds:
*
      WRITE (*, *) ' Enter the number of seconds that is the goal:'
      READ (*, *) goal
*
*  Get lower and upper bounds for n from the standard input device:
*
 1    WRITE (*, *) ' Enter a lower bound for n:'
      READ (*, *, END = 4) n
      IF (n .LE. 0) GO TO 4
      CALL Meter (n, timing, work, ok)
      IF (.NOT. ok) GO TO 1
      IF (timing .GE. goal) THEN
        WRITE(*, *) ' Must take less than', goal, ' sec. Took', timing
        GO TO 1
      END IF
*
 2    WRITE (*, *) ' Enter an upper bound for n:'
      READ (*, *, END = 4) nupper
      IF (nupper .LE. 0) GO TO 4
      CALL Meter (nupper, timing, work, ok)
      IF (.NOT. ok) GO TO 2
      IF (timing .LT. goal) THEN
        WRITE (*, *) ' Must take at least', goal, ' sec. Took', timing
        n = MAX(nupper, n)
        GO TO 2
      END IF
*
*  While the [n, nupper] interval is larger than 1, bisect it and pick a half:
*
 3    IF (nupper - n .GT. 1) THEN
        mean = (n + nupper) / 2
        CALL Meter (mean, timing, work, ok)
        IF (timing .LT. goal) THEN
          n = mean
        ELSE
          nupper = mean
        END IF
        WRITE (*, *) ' New interval: [',n, ',', nupper, ']'
        GO TO 3
      END IF
*
*  Ensure that most recent run was for n, not nupper, and print statistics:
*
      CALL Meter (n, timing, work, ok)
      CALL What (n, timing, work)
 4    END
*
*******************************************************************************
*  This routine should be edited to contain information for your system.      *
*******************************************************************************
      SUBROUTINE What (n, timing, work)
      INTEGER n
      REAL*8 timing, work
*
*  You may change this to single-column output if you need more room.
*
      CHARACTER*72 info(9)
      DATA info /
     &   ' Machine:  SUN 4/370GX          Processor:  SPARC IU+FPU',
     &   ' Memory:   32 MB                # of procs: 1',
     &   ' Cache:    128 KB               # used:     1',
     &   ' NMAX:     512                  Clock:      25 MHz',
     &   ' Disk:     .3 GB SCSI+.7 GB SMD Node name:  tantalus',
     &   ' OS:       Sun OS 4.0.3         Timer:      CPU, etime()',
     &   ' Language: Fortran 77           Alone:      no',
     &   ' Compiler: f77 1.2              Run by:     J. Gustafson',
     &   ' Options:  -O4 -dalign ...      
     &     Date:       17 Oct 1990          '/                          
*
      WRITE (*, *) ' '
      WRITE (*, '(A72)') info
      WRITE (*, 4) work * 1.D-6, timing
 4    FORMAT (' M ops:', F13.6, 12X, 'Time:      ', F6.2, ' seconds')
      WRITE (*, 5) n, (work / timing) * 1.D-6
 5    FORMAT (' n:  ', I9, '                  MFLOPS:', F10.3)
      WRITE (*, 6) 8 * n * n + 120 * n + 800 
 6    FORMAT (' Approximate data memory use:', I10, ' bytes')
      END
*******************************************************************************
*  This routine measures time required on an optical radiosity problem,       *
*  including input, matrix generation, solution, and output.                  *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, 10/17/90.                                                 *
*                                                                             *
*  Calls: Reader  Reads the problem description from secondary storage.       *
*         Region  Partitions box surface into rectangular regions (patches).  *
*         SetUp1  Sets up equations from patch geometries-parallel faces.     *
*         SetUp2  Sets up equations from patch geometries-orthogonal faces.   *
*         SetUp3  Sets up equations-row normalization and radiant properties. *
*         Solver  Solves the equations by LDL factorization.                  *
*         Storer  Stores solution (patch radiosities) on secondary storage.   *
*         Verify  Verifies that radiosities solve equations (not timed)       *
*         When    Returns wall-clock time, in seconds.                        *
*******************************************************************************
      SUBROUTINE Meter (npatch, timing, work, ok)
*
*  Parameters:
*    NMAX    Parameter, largest npatch for your computer; adjust as needed.
*
      PARAMETER (NMAX = 512)
*
*  Passed variables:
*    npatch  In, problem size, here the number of equations (1 per patch).
*    timing  Out, number of seconds measured.
*    work    Out, work done, here the number of floating point operations.
*    ok      Out; .TRUE. if no errors occur, else .FALSE.
*
      INTEGER npatch
      REAL*8 timing, work
      LOGICAL ok
*
*  Local variables:
*    When    Function to return wall clock time in seconds.
*    area    Vector, areas of patches * 8 * pi.
*    box     Vector, dimensions of box in x, y, z directions.
*    coeff   Matrix, the coefficients of the equations to solve.
*    diag    Vectors, diagonal terms of the equations to solve, (R-G-B).
*    emiss   Vector, emissivities of patches in R-G-B components.
*    ops     Vector, floating-point operation counts for routines.
*    p       Vector, number of patches in faces (for operation count).
*    place   Vectors, width-height-depth position of patches.
*    result  Vectors, answer radiosities (R-G-B).
*    rho     Vectors, reflectivities of patches (R-G-B).
*    rhs     Vectors, right-hand sides of equations to solve (R-G-B).
*    sec     Vector, times for routines, in seconds.
*    size    Vectors, width-height sizes of patches.
*    tmp1-2  Floating-point temporary variables.
*    i       General loop counter.
*    iofile  Fortran file number, here set to 9.
*    itmp1   Integer temporary variable.
*    loop    Vectors, patch number ranges for faces.
*    non0    Index of the first nonzero off-diagonal matrix element.
*    tasks   Vector, list of names of routines measured.
*
      REAL*8 When, area(NMAX), box(7), coeff(NMAX, NMAX)
      REAL*8 diag(NMAX, 3), emiss(6, 3), ops(8), p(6)
      REAL*8 place(NMAX, 3), result(NMAX, 3), rho(6, 3)
      REAL*8 rhs(NMAX, 3), sec(8), size(NMAX, 2), tmp1, tmp2
      INTEGER i, iofile, itmp1, loop(6, 2), non0
      CHARACTER*6 tasks(7)
      DATA iofile /9/
      DATA tasks/'Reader', 'Region', 'SetUp1', 'SetUp2',
     &           'SetUp3', 'Solver', 'Storer'/ 
*
*  First check that npatch lies between 6 and NMAX:
*
      IF (npatch .LT. 6) THEN
        WRITE (*, *) ' Must be at least 6, the number of faces.'
        ok = .FALSE.
        RETURN
      ELSE IF (npatch .GT. NMAX) THEN
        WRITE (*, *) ' Exceeds', NMAX, ' = maximum for this system.'
        ok = .FALSE.
        RETURN
      END IF
*
*  Ensure that the 'answer' file does not exist before timing starts;
*
      OPEN (iofile, FILE = 'answer', STATUS = 'UNKNOWN') 
      CLOSE (iofile, STATUS = 'DELETE')
*
*  Time the tasks, individually and collectively (see header descriptions).
*
      sec(1) = When()
      CALL Reader (box, rho, emiss, ok)
      IF (.NOT. ok) RETURN
      sec(2) = When()
      CALL Region (npatch, loop, box, place, size, area, ok)
      IF (.NOT. ok) RETURN
      sec(3) = When()
      CALL SetUp1 (npatch, loop, coeff, place, size)
      sec(4) = When()
      CALL SetUp2 (npatch, loop, coeff, place, size)
      sec(5) = When()
      CALL SetUp3 (npatch, loop, area, rho, emiss, coeff, diag, rhs, ok)
      IF (.NOT. ok) RETURN
      sec(6) = When()
      non0 = loop(2, 1)
      CALL Solver (npatch, non0, coeff, diag, rhs, result)
      sec(7) = When()
      CALL Storer (npatch, loop, place, size, result)
      sec(8) = When()
      timing = sec(8) - sec(1)
      DO 11 i = 1, 7
        sec(i) = sec(i + 1) - sec(i)
 11   CONTINUE
*
*  Assess floating-point work done by each routine called, and total:
*
      itmp1 = 0
      tmp1 = 0.0D0
      DO 12 i = 1, 6
        p(i) = loop(i, 2) - loop(i, 1) + 1
        tmp1 = tmp1 + p(i) ** 2
        itmp1 = itmp1 + INT(SQRT(p(i) * box(i) / box(i + 1)) + .5)
 12   CONTINUE
      tmp2 = p(1) * p(4) + p(2) * p(5) + p(3) * p(6)
      ops(1) = 258.
      ops(2) = 154. + 8. * itmp1 + npatch
      ops(3) = 3. + 840. * tmp2
      ops(4) = 430. * (npatch ** 2 - tmp1 - 2. * tmp2) / 2.
      ops(5) = 72. + 9. * npatch + npatch ** 2 - tmp1
      ops(6) = npatch * (npatch * (npatch + 7.5) - 2.5) - 21.
     &       + non0 * (non0 * (2. * non0 - 16.5) + 35.5)
     &       + non0 * npatch * (9. - 3. * non0)
      ops(7) = 48. * npatch
      work = ops(1) + ops(2) + ops(3) + ops(4)
     &              + ops(5) + ops(6) + ops(7)
*
*  Display timing-work-speed breakdown by routine.
*
      WRITE (*, *) npatch, ' patches:'
      WRITE (*, *)
     & '  Task  Seconds      Operations       MFLOPS    % of Time'
      DO 13 i = 1, 7
        IF (sec(i) .EQ. 0.) sec(i) = .001
        WRITE (*, 14) tasks(i), sec(i), ops(i),
     &       ops(i) / sec(i) / 1.D6, 100. * sec(i) / timing
 13   CONTINUE
      WRITE (*, 14) ' TOTALS: ', timing, work,
     &                           work / timing / 1.D6, 100.
 14   FORMAT (' ', A6, F8.3, F17.0, F14.6, F10.1, ' %')
      CALL Verify (npatch, coeff, diag, rhs, result, ok)
      END
*
*******************************************************************************
*  This function should return the actual, wall clock time (not CPU time),    *
*  as accurately as possible.  Change it to your system timer.                *
*******************************************************************************
      DOUBLE PRECISION FUNCTION When ()
      REAL tdum(2)
*
      When = etime (tdum)
      END
*******************************************************************************
*  The following routine reads in the problem description from secondary      *
*  storage, and checks that numbers are in reasonable ranges.                 *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, 10/17/90.                                                 *
*******************************************************************************
      SUBROUTINE Reader (box, rho, emiss, ok)
*
*  Passed variables:
*    box     Out vector, dimensions of box in x, y, and z directions, repeated
*              to eliminate need for MOD function elsewhere in program.
*    rho     Out vector, reflectivities of interiors of faces, in R-G-B.
*    emiss   Out vector, emissivities of interiors of faces, in R-G-B.
*    ok      Out, .TRUE. if no errors occur, else .FALSE.
*
      REAL*8 box(7), rho(6, 3), emiss(6, 3)
      LOGICAL ok
*
*  Local variables:
*    infile  Device number for input file.
*    i, j    Loop counters.
*    tmp1    Maximum emissivity, to check that emissivities are not all zero.
*
      INTEGER infile, i, j
      REAL*8 tmp1
      DATA infile /9/
*
      OPEN (infile, FILE = 'geom', STATUS = 'OLD', ERR = 103) 
      READ (infile, *, ERR = 103) (box(i), i = 1, 3)
      READ (infile, *, ERR = 103) ((rho(i, j), i = 1, 6), j = 1, 3)
      READ (infile, *, ERR = 103) ((emiss(i, j), i = 1, 6), j = 1, 3)
C      CLOSE (infile) Bodin
*
      DO 102 j = 1, 3
        IF (box(j) .LT. 1. .OR. box(j) .GE. 100.) THEN
          WRITE (*, *) ' Box dimensions must be between 1 and 100.'
          ok = .FALSE.
          RETURN
        END IF
        box(j + 3) = box(j)
*
        tmp1 = 0.
        DO 101 i = 1, 6
          IF (rho(i, j) .LT. 0.001D0 .OR. rho(i, j) .GT. 0.999D0) THEN
            WRITE (*, *) ' Reflectivities must be between .001 and .999'
            ok = .FALSE.
            RETURN
          END IF
          IF (emiss(i, j) .LT. 0.) THEN
            WRITE (*, *) ' Emissivity cannot be negative.'
            ok = .FALSE.
            RETURN
          END IF
          IF (tmp1 .LT. emiss(i, j)) tmp1 = emiss(i, j)
 101    CONTINUE
        IF (tmp1 .EQ. 0.) THEN
          WRITE (*, *) ' Emissivities are zero.  Problem is trivial.'
          ok = .FALSE.
          RETURN
        END IF
 102  CONTINUE
      box(7) = box(4)
      ok = .TRUE.
      RETURN
*
*  Trap if 'geom' file is missing or otherwise isn't read correctly:
*
 103  WRITE (*, *) ' Problem reading geom file.'
      ok = .FALSE.
      END
*******************************************************************************
*  The following routine decomposes the surface of a variable-sized box       *
*  into patches that are as nearly equal in size and square as possible.      *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, 10/17/90.                                                 *
*******************************************************************************
      SUBROUTINE Region (npatch, loop, box, place, size, area, ok)
*
*  Passed variables:
*    loop    Out vectors, patch number ranges for faces.
*    npatch  In, problem size.
*    area    Out vector, 8 * pi * areas of the patches.
*    box     In vector, the x, y, z dimensions of the box.
*    place   Out vectors, width-height-depth positions of patches.
*    size    Out vectors, width-height sizes of patches.
*    ok      Out flag, .TRUE. if no errors, else .FALSE.
*
      INTEGER loop(6, 2), npatch
      REAL*8 area(npatch), box(7)
      REAL*8 place(npatch, 3), size(npatch, 2)
      LOGICAL ok
*
*  Local variables:
*    icol    Loop counter over the number of columns.
*    iface   Loop counter over the number of faces.
*    ipatch  Loop counter over the number of patches.
*    itmp1-2 Integer temporary variables.
*    last    Inner loop ending value.
*    lead    Inner loop starting value.
*    numcol  Number of columns on faces.
*    numpat  Number of patches on a face.
*    numrow  Number of rows of patches in a column.
*    height  Height of a patch within a column.
*    tmp1-4  Floating-point temporary variables.
*    width   Width of a column of patches.
*
      INTEGER icol, iface, ipatch, itmp1, itmp2
      INTEGER last, lead, numcol, numpat, numrow
      REAL*8 height, tmp1, tmp2, tmp3, tmp4, width
*
*  Allocate patches to each face, proportionate to area of each face:
*
      tmp1 = 2. * (box(1) * box(2) + box(2) * box(3) + box(3) * box(1))
      tmp2 = 0.D0
      tmp3 = npatch
      loop(1, 1) = 1
      DO 201 iface = 1, 5
        tmp2 = tmp2 + box(iface) * box(iface + 1)
        loop(iface, 2) = INT(tmp3 * tmp2 / tmp1 + .5)
        loop(iface + 1, 1) = loop(iface, 2) + 1
 201  CONTINUE
      loop(6, 2) = npatch
*
*  Subdivide each face into numpat patches:
*
      DO 204 iface = 1, 6
        numpat = loop(iface, 2) - loop(iface, 1) + 1
        tmp3 = 0.D0
        IF (iface .GT. 3) tmp3 = box(iface - 1)
        numcol = INT(SQRT(numpat * box(iface) / box(iface + 1)) + .5)
        IF (numcol .GT. numpat) numcol = numpat
        IF (numcol .EQ. 0) numcol = 1
        width = box(iface) / numcol
        itmp1 = numcol - 1
        tmp1 = 0.D0
        DO 203 icol = 1, numcol
          itmp2 = itmp1 / numcol
          numrow = (itmp1 + numpat) / numcol - itmp2
          IF (numrow .EQ. 0) THEN
            WRITE(*, *) ' Eccentric box requires more patches.'
            ok = .FALSE.
            RETURN
          END IF
          height = box(iface + 1) / numrow
          tmp2 = 0.D0
          tmp4 = width * height * (8.D0 * 3.14159265358979323D0)
          lead = loop(iface, 1) + itmp2
          last = lead + numrow - 1
          DO 202 ipatch = lead, last
            size(ipatch, 1) = width
            size(ipatch, 2) = height
            place(ipatch, 1) = tmp1
            place(ipatch, 2) = tmp2
            place(ipatch, 3) = tmp3
            area(ipatch) = tmp4
            tmp2 = tmp2 + height
 202      CONTINUE
          tmp1 = tmp1 + width
          itmp1 = itmp1 + numpat
 203    CONTINUE
 204  CONTINUE
*
      ok = .TRUE.
      END
*******************************************************************************
*  This routine sets up the radiosity matrix for parallel patches.            *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, 10/17/90.                                                 *
*******************************************************************************
      SUBROUTINE SetUp1 (npatch, loop, coeff, place, size)
*
*  Passed variables:
*    npatch  In, the problem size.
*    loop    In vectors, patch number ranges for faces.
*    coeff   Out matrix, the coefficients of the system to solve.
*    place   In vectors, width-height-depth positions of patches.
*    size    In vectors, width-height sizes of patches.
*
      INTEGER loop(6, 2)
      REAL*8 coeff(npatch, npatch)
      REAL*8 place(npatch, 3), size(npatch, 2)
*
*  Local variables:
*    i-n     General loop counters.
*    iface   Loop counter over the number of faces.
*    ipatch  Loop counter over the number of patches.
*    jface   Face coupled to iface when computing matrix elements.
*    jpatch  Patch coupled to ipatch when computing matrix elements.
*    d       Array for point-to-point couplings between patch corners.
*    d2      Array of squares of d values, to save recomputation.
*    tmp1-8  Floating-point temporary variables.
*
      INTEGER i, j, k, m, n, iface, ipatch, jface, jpatch
      REAL*8 d(2, 2, 2), d2(2, 2, 2)
      REAL*8 tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7, tmp8
*
      DO 309 iface = 1, 3
        jface = iface + 3
        tmp1 = place(loop(jface, 1), 3) ** 2
        DO 308 ipatch = loop(iface, 1), loop(iface, 2)
          DO 307 jpatch = loop(jface, 1), loop(jface, 2)
            DO 301 j = 1, 2
              d(1, 1, j) = place(jpatch, j) - place(ipatch, j)
              d(2, 1, j) = d(1, 1, j) + size(jpatch, j)
              d(1, 2, j) = d(1, 1, j) - size(ipatch, j)
              d(2, 2, j) = d(2, 1, j) - size(ipatch, j)
              d2(1, 1, j) = d(1, 1, j) ** 2
              d2(2, 1, j) = d(2, 1, j) ** 2
              d2(1, 2, j) = d(1, 2, j) ** 2
              d2(2, 2, j) = d(2, 2, j) ** 2
 301        CONTINUE
            tmp2 = 0.0
            DO 306 j = 1, 2
              DO 305 m = 1, 2
                DO 304 i = 1, 2
                  tmp3 = d2(m, i, 3 - j) + tmp1
                  tmp4 = SQRT(tmp3)
                  tmp5 = 1.D0 / tmp4
                  tmp8 = 0.D0
                  DO 303 k = 1, 2
                    DO 302 n = 1, 2
                      tmp6 = d2(k, n, j)
                      tmp7 = d(k, n, j)
                      tmp2 = (tmp3 - tmp6) * LOG(tmp6 + tmp3) - tmp2
                      tmp8 = -tmp7 * ATAN(tmp7 * tmp5) - tmp8
 302                CONTINUE
                    tmp2 = -tmp2
                    tmp8 = -tmp8
 303              CONTINUE
                  tmp2 = -4.D0 * tmp4 * tmp8 - tmp2
 304            CONTINUE
                tmp2 = -tmp2
 305          CONTINUE
 306        CONTINUE
            coeff(ipatch, jpatch) = tmp2
            coeff(jpatch, ipatch) = tmp2
 307      CONTINUE
 308    CONTINUE
 309  CONTINUE
      END
*******************************************************************************
*  This routine sets up the radiosity matrix for orthogonal patches.          *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, 10/17/90.                                                 *
*******************************************************************************
      SUBROUTINE SetUp2 (npatch, loop, coeff, place, size)
*
*  Passed variables:
*    npatch  In, the problem size.
*    loop    In vectors, patch number range for faces.
*    coeff   Out matrix, the coefficients of the system to solve.
*    place   In vectors, width-height-depth positions of patches.
*    size    In vectors, width-height sizes of patches.
*
      INTEGER loop(6, 2)
      REAL*8 coeff(npatch, npatch)
      REAL*8 place(npatch, 3), size(npatch, 2)
*
*  Local variables:
*    i-n     General loop counters.
*    iface   Loop counter over the number of faces.
*    ipatch  Loop counter over the number of patches.
*    jface   Face coupled to iface when computing matrix elements.
*    jpatch  Patch coupled to ipatch when computing matrix elements.
*    r, s    Arrays for point-to-point couplings between patch corners.
*    r2      Array of squares of r values, to save recomputation.
*    tmp2-8  Floating-point temporary variables.
*
      INTEGER i, iface, ipatch, j, jface, jpatch, k, m, n
      REAL*8 r(2, 2), r2(2, 2), s(2, 2)
      REAL*8 tmp2, tmp3, tmp4, tmp5, tmp6, tmp7, tmp8
*
      DO 410 iface = 1, 6
        DO 409 m = 1, 2
          jface = MOD(iface + m - 1, 6) + 1
          DO 408 ipatch = loop(iface, 1), loop(iface, 2)
            DO 407 jpatch = loop(jface, 1), loop(jface, 2)
              r(1, 1) = place(jpatch, m) - place(ipatch, 3 - m)
              r(2, 1) = r(1, 1) + size(jpatch, m    )
              r(1, 2) = r(1, 1) - size(ipatch, 3 - m)
              r(2, 2) = r(2, 1) - size(ipatch, 3 - m)
              s(1, 1) = place(ipatch, m    ) - place(jpatch, 3)
              s(2, 1) = s(1, 1) + size(ipatch, m    )
              s(1, 2) = place(jpatch, 3 - m) - place(ipatch, 3)
              s(2, 2) = s(1, 2) + size(jpatch, 3 - m)
              DO 402 i = 1, 2
                DO 401 j = 1, 2
                  s(i, j) = s(i, j) ** 2
                  r2(i, j) = r(i, j) ** 2
 401            CONTINUE
*
*  Bump the term by a small real to avoid singularities in coupling function:
*
                s(i, 2) = s(i, 2) + 1.D-35
 402          CONTINUE
              tmp2 = 0.0
              DO 406 i = 1, 2
                DO 405 j = 1, 2
                  tmp3 = s(i, 1) + s(j, 2)
                  tmp4 = SQRT(tmp3)
                  tmp5 = 1.D0 / tmp4
                  tmp8 = 0.0
                  DO 404 k = 1, 2
                    DO 403 n = 1, 2
                      tmp6 = r2(k, n)
                      tmp7 = r(k, n)
                      tmp2 = (tmp3 - tmp6) * LOG(tmp6 + tmp3) - tmp2
                      tmp8 = -tmp7 * ATAN(tmp7 * tmp5) - tmp8
 403                CONTINUE
                    tmp2 = -tmp2
                    tmp8 = -tmp8
 404              CONTINUE
                  tmp2 = -4.D0 * tmp4 * tmp8 - tmp2
 405            CONTINUE
                tmp2 = -tmp2
 406          CONTINUE
              tmp2 = ABS(tmp2)
              coeff(ipatch, jpatch) = tmp2
              coeff(jpatch, ipatch) = tmp2
 407        CONTINUE
 408      CONTINUE
 409    CONTINUE
 410  CONTINUE
      END
*******************************************************************************
*  This routine sets up the radiosity matrix... normalizes row sums to 1,     *
*  and includes terms derived from reflectivites and emissivities of faces.   *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, 10/17/90.                                                 *
*******************************************************************************
      SUBROUTINE SetUp3 (npatch, loop, area, rho, emiss, coeff, diag,
     &                   rhs, ok)
*
*  Passed variables:
*    npatch  In, the problem size.
*    loop    In vectors, patch number ranges of the faces.
*    area    In vector, 8 * pi * areas of the patches.
*    rho     In vectors, reflectivities of the face interiors (R-G-B).
*    emiss   In vectors, emissivities of the face interiors (R-G-B).
*    coeff   Out matrix, coefficients of the system to solve.
*    diag    Out vectors, diagonal terms of the system to solve (R-G-B).
*    rhs     Out vectors, right-hand sides of system to solve (R-G-B).
*    ok      Out, .TRUE. if no errors occur, else .FALSE.
*
      INTEGER npatch, loop(6, 2)
      REAL*8 area(npatch), emiss(6, 3), coeff(npatch, npatch)
      REAL*8 diag(npatch, 3), rho(6, 3), rhs(npatch, 3)
      LOGICAL ok
*
*  Local variables:
*    iface     Loop counter over the number of faces.
*    ipatch    Inner loop counter over the number of patches.
*    j         Loop counter over each color (R-G-B).
*    jpatch    Outer loop counter over the number of patches.
*    tmp1      Floating-point temporary variable.
*    vtmp1-2   Floating-point vector temporary variables.
*
      INTEGER iface, ipatch, j, jpatch
      REAL*8 tmp1, vtmp1(3), vtmp2(3)
*
*  Ensure that row sums to 1, and put in reflectivities (rho) and emissivities:
*
      DO 507 iface = 1, 6
        DO 501 j = 1, 3
          vtmp1(j) = 1.D0 / rho(iface, j)
          vtmp2(j) = emiss(iface, j) * vtmp1(j)
 501    CONTINUE
        DO 506 jpatch = loop(iface, 1), loop(iface, 2)
          tmp1 = 0.D0
          DO 502 ipatch = 1, loop(iface, 1) - 1
            tmp1 = tmp1 + coeff(ipatch, jpatch)
 502      CONTINUE
          DO 503 ipatch = loop(iface, 2) + 1, npatch
            tmp1 = tmp1 + coeff(ipatch, jpatch)
 503      CONTINUE
*
*  Make sure row sum (total form factor) is close to 1:
*
          IF (ABS(tmp1 - area(jpatch)) .GT. (0.5D-8 * tmp1)) THEN
            WRITE (*, *) ' Total form factor is too far from unity.'
            ok = .FALSE.
            RETURN
          END IF
          tmp1 = -tmp1
*
*  Set coplanar patch interactions to zero:
*
          DO 504 ipatch = loop(iface, 1), loop(iface, 2)
            coeff(ipatch, jpatch) = 0.D0
 504      CONTINUE
*
*  Assign diagonal entries and right-hand sides:
*
          DO 505 j = 1, 3
            diag(jpatch, j) = vtmp1(j) * tmp1
            rhs(jpatch, j) = vtmp2(j) * tmp1
 505      CONTINUE
 506    CONTINUE
 507  CONTINUE
      ok = .TRUE.
      END
*******************************************************************************
*  This routine factors and backsolves a real, symmetric, near-dense matrix   *
*  by LDL factorization.  No pivoting; the matrix is diagonally dominant.     *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, 10/17/90.                                                 *
*******************************************************************************
      SUBROUTINE Solver (npatch, non0, coeff, diag, rhs, result)
*
*  Passed variables:
*    non0    In, index of first nonzero off-diagonal matrix element.
*    npatch  In, number of equations to solve.
*    coeff   In/out matrix; coefficients of system to solve.
*    diag    In/temporary vectors; diagonal coefficients of system (R-G-B).
*    rhs     In vectors; right-hand sides from emissivities (R-G-B).
*    result  Out vectors; solution radiosities (R-G-B).
*
      INTEGER non0, npatch
      REAL*8 coeff(npatch, npatch), diag(npatch, 3)
      REAL*8 rhs(npatch, 3), result(npatch, 3)
*
*  Local variables:
*    tmp1    Floating-point temporary variable.
*    i-m     General loop counters.
*
      REAL*8 tmp1
      INTEGER i, j, k, m
*
*  Load lower triangle of coefficients, diagonal, and solution vector:
*
      DO 615 m = 1, 3
        DO 602 i = non0, npatch
          coeff(i, i) = diag(i, m)
          result(i, m) = rhs(i, m)
          DO 601 j = 1, i - 1
            coeff(j, i) = coeff(i, j)
 601      CONTINUE
 602    CONTINUE
*
*  Factor matrix, writing factors on top of original matrix.
*
        DO 603 j = 1, non0 - 1
          coeff(j, j) = 1.D0 / diag(j, m)
          result(j, m) = rhs(j, m)
 603    CONTINUE
        DO 607 j = non0, npatch
          DO 605 k = non0, j - 1
            DO 604 i = 1, k - 1
               coeff(k, j) = coeff(k, j) - (coeff(i, k) * coeff(i, j))
 604        CONTINUE
 605      CONTINUE
          DO 606 k = 1, j - 1
            tmp1 = coeff(k, j)
            coeff(k, j) = tmp1 * coeff(k, k)
            coeff(j, j) = coeff(j, j) - tmp1 * coeff(k, j)
 606      CONTINUE
          coeff(j, j) = 1.D0 / coeff(j, j)
 607    CONTINUE
*
*  Backsolve, in three stages (for L, D, and L transpose):
*
        DO 609 j = non0, npatch
          DO 608 i = 1, j - 1
            result(j, m) = result(j, m) - (result(i, m) * coeff(i, j))
 608      CONTINUE
 609    CONTINUE
*
        DO 610 j = 1, npatch
          result(j, m) = result(j, m) * coeff(j, j)
 610    continue
*
        DO 612 j = npatch - 1, non0, -1
          DO 611 i = j + 1, npatch
            result(j, m) = result(j, m) - (result(i, m) * coeff(j, i))
 611      CONTINUE
 612    CONTINUE
*
        DO 614 j = non0 - 1, 1, -1
          DO 613 i = non0, npatch
            result(j, m) = result(j, m) - (result(i, m) * coeff(j, i))
 613      CONTINUE
 614    CONTINUE
*
 615  CONTINUE
      END
*******************************************************************************
*  The following routine writes the answer to secondary storage.              *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, 10/17/90.                                                 *
*******************************************************************************
      SUBROUTINE Storer (npatch, loop, place, size, result)
*
*  Passed variables:
*    npatch  In, problem size (number of patches).
*    loop    In vectors, patch number ranges for faces.
*    place   In vectors, width-height-depth positions of patches.
*    size    In vectors, width-height sizes of patches.
*    result  In vectors, solution radiosities of patches (R-G-B).
*
      INTEGER npatch, loop(6, 2)
      REAL*8 place(npatch, 3), size(npatch, 2), result(npatch, 3)
*
*  Local variables:
*    i       General loop counter.
*    iface   Loop counter over number of faces.
*    iofile  Device number for output file.
*    ipatch  Loop counter of number of patches within a face.
*
      INTEGER i, iface, iofile, ipatch
      DATA iofile /10/
*
*  Write patch geometry to 'answer' file:
*
      OPEN (iofile, FILE = 'answer', STATUS = 'UNKNOWN', ERR = 707)
      WRITE (iofile, *, ERR = 707) npatch, ' patches:'
      WRITE (iofile, *, ERR = 707) '  Patch  Face       ',
     &       'Position in w, h, d            Width     Height'
      DO 702 iface = 1, 6
        DO 701 ipatch = loop(iface, 1), loop(iface, 2)
          WRITE (iofile, 703, ERR = 707) ipatch, iface
          WRITE (iofile, 703, ERR = 707)  (place(ipatch, i), i = 1, 3)
          WRITE (iofile, 703, ERR = 707)  (size(ipatch, i), i = 1, 2)
 701    CONTINUE
 702  CONTINUE
 703  FORMAT (' ', I5, '   ', I4, 3F11.5, '  ', 2F11.5)
*
*  Write patch radiosities to 'answer' file:
*
      WRITE (iofile, *, ERR = 707) ' '
      WRITE (iofile, *, ERR = 707) '  Patch  Face    Radiosities'
      DO 705 iface = 1, 6
        DO 704 ipatch = loop(iface, 1), loop(iface, 2)
          WRITE (iofile, 706, ERR = 707) ipatch, iface
          WRITE (iofile, 706, ERR = 707)  (result(ipatch, i), i = 1, 3)
 704    CONTINUE
 705  CONTINUE
 706  FORMAT (' ', I5, '   ', I4, 3F12.8)
*
      CLOSE (iofile)
      RETURN
*
*  Error trap for output problems, like insufficient mass storage:
*
 707  WRITE (*, *) ' Cant write answer file.'
      END
*******************************************************************************
*  This routine verifies that the computed radiosities satisfy the equations. *
*                                                                             *
*  John Gustafson, Diane Rover, Stephen Elbert, and Michael Carter;           *
*  Ames Laboratory, 10/17/90.                                                 *
*******************************************************************************
      SUBROUTINE Verify (npatch, coeff, diag, rhs, result, ok)
*
*  Passed variables:
*    npatch  In, number of equations to solve.
*    coeff   In matrix; coefficients of system to solve.
*    diag    In vectors; diagonal coefficients of system to solve (R-G-B).
*    rhs     In vectors; right-hand sides of linear systems to solve (R-G-B).
*    result  In vectors; solution radiosities (R-G-B).
*    ok      Out; flag indicating success or failure.
*
      INTEGER npatch
      REAL*8 coeff(npatch, npatch), diag(npatch, 3)
      REAL*8 rhs(npatch, 3), result(npatch, 3)
      LOGICAL ok
*
*  Local variables:
*    anorm   Norm of coeff matrix, max of absolute values of its elements.
*    tmp1-2  Floating-point temporary variables.
*    xnorm   Norm of result vector, max of absolute values of its elements.
*    i, j    General loop counters.
*
      REAL*8 anorm, tmp1, tmp2, xnorm
      INTEGER i, j
*
*  Copy lower triangle of coefficients to upper triangle, and load diagonal:
*
      DO 905 m = 1, 3
        tmp1 = 0.D0
        DO 902 i = 1, npatch
          coeff(i, i) = diag(i, m)
          DO 901 j = 1, i - 1
            coeff(j, i) = coeff(i, j)
 901      CONTINUE
 902    CONTINUE
*
*  Multiply matrix by solution vector, and accumulate norm of residual:
*
        anorm = 0.D0
        xnorm = 0.D0
        DO 904 j = 1, npatch
          tmp2 = rhs(j, m)
          DO 903 i = 1, npatch
            tmp2 = tmp2 - coeff(i, j) * result(i, m)
            anorm = MAX(anorm, ABS(coeff(i, j)))
 903      CONTINUE
          xnorm = MAX(xnorm, ABS(result(j, m)))
          tmp1 = tmp1 + ABS(tmp2)
 904    CONTINUE
*
        tmp1 = tmp1 / (anorm * xnorm)
        IF (tmp1 .GT. 0.5D-8) THEN
          WRITE (*, *) ' Residual is too large: ', tmp1
          ok = .FALSE.
        ELSE
          ok = .TRUE.
        END IF
 905  CONTINUE
      END
